/* Dynamic classes
   Bob Gray
   Gisli Hjalmtysson
   Network Mathematics Research Department (AK0112730), AT&T 
   600 Mountain Avenue
   Murray Hill, New Jersey

   4 June 1996

   dynmedium.h

   This file defines the template class that provides for versioning.
*/

#ifdef _WIN32
#define THROW1(e1)
#define THROW2(e1,e2)
#define THROW3(e1,e2,e3)
#define THROW4(e1,e2,e3,e4)
#define THROW5(e1,e2,e3,e4,e5)
#else
#define THROW1(e1)              throw(e1)
#define THROW2(e1,e2)           throw(e1,e2)
#define THROW3(e1,e2,e3)        throw(e1,e2,e3)
#define THROW4(e1,e2,e3,e4)     throw(e1,e2,e3,e4)
#define THROW5(e1,e2,e3,e4,e5)  throw(e1,e2,e3,e4,e5)
#endif

#ifndef _DYNAMIC_H
#define _DYNAMIC_H

#include "lock.h"
#include <assert.h>

#ifdef _WIN32
// no equivalents
#else
#include <dirent.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "boolean.h"
#include "dynexcept.h"
#include "dynversion.h"
#include "faststring.h"
#include "genHash.h"
#include "list.h"
#include "load.h"
#include "manager.h"
#include "map.h"

#ifdef ALLOW_DOWNLOADS
#ifdef _WIN32
#include "win_download.h"
#else
#include "download.h"
#endif
#endif

    /* the main proxy */

typedef list<DynamicVersion*> VersionList;

template<class T> class dynmedium {

    protected:

	    /*    
	     * the map of class names to handles, the map of handles to a
             * version list, the map of handles to the active version, the
	     * map of handles to class names, the current number of class
             * names, and a lock for mutual exclusion
	     */
	
	static Hash *handleMap;
	static VersionList **versionMap;
	static DynamicVersion **activeMap;
	static char **nameMap;
	static int numClasses;
#ifdef USE_THREADS
	static MutexLock lock;
#endif

	    /*
	     * The actual object, its version information and its class handle
	     */

	T *object;
	DynamicVersion *dvPtr;
	int classHandle;

	    /* 
	     * activation 
	     */

	static int activateInternal
	    (Boolean invalidate, 
	     const char *actualName, 
	     const char *className = NULL)
	    THROW5 (NoTemporaryLibrary,NoDownload,NoVersionLibrary,InvalidVersion,NoCreateFunction);

	    /* deactivation */

	static void eraseVersion
	    (int handle, DynamicVersion *ptr); 

	static void deactivateVersion
	    (int handle, DynamicVersion *ptr, Boolean force);

	static void invalidateAll
	    (int handle, const char *actualName);

    public:

	    /* constructor, copy constructor and destructor */ 

        dynmedium (int handle) 
	    THROW2 (InvalidHandle,NoObject);

	dynmedium (const dynmedium<T>& proxy)
	    THROW2 (NoObject,ZombieVersion);

       ~dynmedium ();

	    /* assignment operator */

	dynmedium<T>& operator= (const dynmedium<T>& proxy)
	    THROW1 (ZombieVersion);

	    /* smart pointer to reference the actual object */

	T *operator-> (void) 
	    THROW1 (ZombieVersion);

	    /* activate a version */

	static int activate 
	    (const char *actualName, const char *className = NULL)
	    THROW5 (NoTemporaryLibrary,NoDownload,NoVersionLibrary,InvalidVersion,NoCreateFunction);
 
	    /* invalidate a version */

	static void invalidate 
	    (const char *actualName, int handle)
	    THROW3 (InvalidHandle,ActiveVersion,NonexistentVersion);

	    /* invalidate current versions and activate a new version */

	static int activate_and_invalidate 
	    (const char *actualName, const char *className = NULL)
	    THROW5 (NoTemporaryLibrary,NoDownload,NoVersionLibrary,InvalidVersion,NoCreateFunction);

	    /* information routines -- need more routines */ 

	static int getVersionCount 
	    (int handle);
};

/* static data */

template <class T>
    Hash * dynmedium<T>::handleMap = NULL;

template <class T>
    VersionList **dynmedium<T>::versionMap = NULL;

template <class T>
    DynamicVersion **dynmedium<T>::activeMap = NULL;

template <class T>
    char **dynmedium<T>::nameMap = NULL;

template <class T>
    int dynmedium<T>::numClasses = 0;

#ifdef USE_THREADS

template <class T>
    MutexLock dynmedium<T>::lock;

#endif

/* template<class T> dynmedium<T>::dynmedium

   Purpose: This procedure is the constructor for class "dynamic".

     Input: handle = handle of the class for which we are creating the instance
		     (int)

    Output: The procedure initializes this instance of "dynamic", i.e., it
            creates an object of the given class and adds this object
   	    to the internal control structures.
*/

template<class T> dynmedium<T>::dynmedium (int handle) THROW2(InvalidHandle,NoObject)
{
	/* lock for mutual exclusion -- automatically released */

    DEFINE_GUARD(guard,lock);

	/* check the handle */

    if ((handle < 0) || (handle > numClasses) || (activeMap[handle] == NULL)) {
	throw InvalidHandle();
    }

	/* start out empty */

    object      = NULL;
    dvPtr       = activeMap[handle];
    classHandle = handle;

	/* call the function to create the object -- we know that funcPtr  */
	/* is not NULL since a version can never become the active version */
	/* unless activate was able to load the library and find the       */
	/* createInstance method                                           */

    object = (T *)(*(dvPtr->funcPtr))(); 

    if (object == NULL) {
	throw NoObject ();
    }

	/* increment the reference count for the library */ 

    dvPtr -> refCount += 1;
}

/* template <class T> dynmedium<T>::dynmedium

   Purpose: This procedure is the copy constructor for class "dynamic".

     Input: proxy = the proxy that is being copied
		    (const class dynmedium<T>&)

    Output: The procedure copies the object.
*/

template<class T> dynmedium<T>::dynmedium (const dynmedium<T>& proxy) THROW2 (NoObject,ZombieVersion)
{
	/* lock the list that contains the *copied* object's version */

    DEFINE_GUARD(guard,proxy.lock);

	/* disallow the copy if we have a zombie version */ 

    if (proxy.dvPtr -> force) {
	throw ZombieVersion (); 
    } 

	/* copy the object */

    if ((object = (T *) proxy.object -> copyInstance()) == NULL) {
	throw NoObject();
    }

	/* reference the appropriate version and increment reference count */

    dvPtr = proxy.dvPtr;
    dvPtr -> refCount += 1;
}

/* template <class T> dynmedium<T>::~dynmedium

   Purpose: This procedure is the destructor for class "dynamic".

     Input: None

    Output: The procedure destroys this instance of "dynamic", i.e., it
            destroys the associated object of interface type T and removes 
	    this object from the internal control structures.
*/

template<class T> void dynmedium<T>::eraseVersion (int handle, DynamicVersion *dvPtr)
{
    versionMap[handle] -> remove (dvPtr);

    if (dvPtr == activeMap[handle]) {
	activeMap[handle] = NULL;
    }

    unloadLibrary (dvPtr);
    delete (dvPtr);
}

template<class T> dynmedium<T>::~dynmedium() 
{
    list<DynamicVersion*>::iterator i;

	/* assertions */

    assert (dvPtr != NULL);
    assert (object != NULL);
    assert (dvPtr -> refCount > 0);

	/* lock for mutual exclusion -- automatically released */

    DEFINE_GUARD(guard,lock);

	/* delete the actual object -- this must be done FIRST since */
	/* we need the destructor, i.e., we can't do the "dlclose"   */
	/* until we are finished with the destructor                 */

    delete (object);

	/* decrement the *version* reference count and remove the */
	/* version if we are done with it			  */

    dvPtr -> refCount -= 1;

    if ((dvPtr -> refCount == 0) && !(dvPtr -> active)) {
	eraseVersion (classHandle, dvPtr);
    }
} 

/* template <class T> dynmedium<T>::operator=

   Purpose: This procedure is the assignment operator for class dynmedium<T>.

     Input: proxy = proxy from which we are assigning
		    (const class dynmedium<T>&)
*/

template <class T> dynmedium<T> & dynmedium<T>::operator= (const dynmedium<T>& proxy) THROW1(ZombieVersion)
{
    T *oldObject;
    DynamicVersion *oldDvPtr;

	/* locks for mutual exclusion -- automatically released */

    DEFINE_GUARD(guard,lock);
    DEFINE_GUARD(otherGuard,proxy.lock);
	
	/* disallow the assignment if we have a zombie version */
	/* or incompatible versions                            */

    if (proxy.dvPtr -> force) {
	throw ZombieVersion();
    }

	/* nothing to do if we are the same proxy */

    if (this == &proxy) {
	return (*this);
    }

	/* nothing to do if we refer to the same object */

    if (object == proxy.object) {
	return (*this);
    }

	/* remember the version information that we're about to replace */
 
    oldDvPtr = dvPtr;

	/* do the assignment */

    if (dvPtr -> isSame (*proxy.dvPtr)) {

        proxy.object -> assignInstance (*object);

    } else {

	oldObject = object;

	if ((object = (T *) proxy.object -> copyInstance()) == NULL) {
	    object = oldObject;
	    throw NoObject ();
	}

	delete (oldObject);

    } 

    dvPtr = proxy.dvPtr;
    dvPtr -> refCount += 1;

	/* see if we are done with the old version */

    oldDvPtr -> refCount -= 1;

    if ((oldDvPtr -> refCount == 0) && !(oldDvPtr -> active)) {
	eraseVersion (classHandle, oldDvPtr);
    } 

	/* done */

    return (*this);
}

/* template <class T> dynmedium<T>::activate

   Purpose: Register a class version as the current active version 

     Input: actualName = actual name of library that contains the desired class
			 version 
		         (const char *) 

	    className  = name of the class 
			 (const char *)

    Output: The procedure marks the class version as the current version
	    and returns the HANDLE of the class.
 
      Note: className has a default value of NULL.  If className is NULL,
	    the function uses the default class name associated with 
            interface class T. 
*/

const int DEFAULT_HANDLES = 5;

template<class T> int dynmedium<T>::activateInternal (Boolean invalidate, const char *actualName, const char *className) THROW5(NoTemporaryLibrary,NoDownload,NoVersionLibrary,InvalidVersion,NoCreateFunction)
{
    int handle;
    Boolean file; 
    const char *localName;
#ifdef ALLOW_DOWNLOADS
    char *temporaryPath;
    HeapPtr<char> temporaryName;
#endif
    VersionList::iterator i;

	/* assertions */

    assert (actualName != NULL);

	/* get the classname */

    if (className == NULL) {
	className = T::getClassName ();
    }

	/* lock for mutual exclusion -- automatically released */

    DEFINE_GUARD(guard,lock);

	/* allocate the version structures if necessary */

    if (handleMap == NULL) {

	handleMap  = new Hash ();
	versionMap = new VersionList * [DEFAULT_HANDLES];
	activeMap  = new DynamicVersion * [DEFAULT_HANDLES];
	nameMap    = new char * [DEFAULT_HANDLES];
	numClasses = DEFAULT_HANDLES;

	for (int n = 0; n < numClasses; n++) {
	    versionMap[n] = NULL;
	    activeMap [n] = NULL;
	    nameMap   [n] = NULL;
	}
    }

	/* get a handle for the given class */

    if (handleMap -> lookup (className, handle) < 0) {

	for (handle = 0; handle < numClasses; handle++) { 
	    if (versionMap[handle] == NULL) {
		versionMap[handle] = new VersionList;
	        nameMap   [handle] = my_strcpy (className);
		activeMap [handle] = NULL;
		handleMap -> add (className, handle);
		break;
	    }
	}

	if (handle == numClasses) {

	    int newNumClasses = numClasses << 1;
	    VersionList **newVersionMap = new VersionList * [newNumClasses];
	    char **newNameMap = new char * [newNumClasses];
	    DynamicVersion **newActiveMap = new DynamicVersion * [newNumClasses];

	    for (int i = 0; i < numClasses; i++) {
	 	newVersionMap[i] = versionMap[i];
	        newNameMap   [i] = nameMap   [i];
	        newActiveMap [i] = activeMap [i];
	    }

	    delete (versionMap);
	    delete (nameMap);
	    delete (activeMap);
	    versionMap = newVersionMap;
	    nameMap = newNameMap;
	    activeMap = newActiveMap;
	    numClasses = newNumClasses;

	    versionMap[handle] = new VersionList;
	    nameMap   [handle] = my_strcpy (className);
	    activeMap [handle] = NULL;
	    handleMap -> add (className, handle);
	} 
    }

	/* see if the new version already exists */

    DynamicVersion *ptr = NULL;
    VersionList &versionList = *versionMap[handle];
     
    for (i = versionList.begin(); i != versionList.end(); ++i) {
	if ((*i) -> isLibrary (actualName)) {
	    ptr = (*i);
	    break;
	}
    }

    if (ptr != NULL) {

	    /* invalidate all non-active versions if we are invalidating; */
	    /* otherwise just unmark the current active version if it is  */
	    /* not the same as the new active version                     */  

	if (invalidate == e_TRUE) {

	    invalidateAll (handle, actualName);

 	} else if (ptr != activeMap[handle]) {

	    deactivateVersion (handle, activeMap[handle], e_FALSE);
	}

	    /* turn on the newly activated version */

	ptr -> active     = e_TRUE;
	ptr -> force      = e_FALSE;
	activeMap[handle] = ptr;
	return (handle);
    }

	/* If we are allowing downloads, we might have to get the version */
	/* from the World Wide Web.  Otherwise the version always comes   */
	/* from the local filesystem.                                     */

#ifdef ALLOW_DOWNLOADS

    if (!strncmp (actualName, "http://", 7)) {

	if ((temporaryPath = getenv("TEMPORARY_LIBRARY")) == NULL) {
	    throw NoTemporaryLibrary ((char *) NULL);
	}

	if ((temporaryName = downloadVersion (actualName, temporaryPath)) == NULL) {
	    throw NoDownload (actualName);
	}

	file = e_FALSE;
        localName = temporaryName;

    } else {

	file = e_TRUE;
	localName = actualName;
    }

#else

    localName = actualName;
    file = e_TRUE;

#endif   /* ALLOW_DOWNLOADS */

	/* remember the version information */

    DynamicVersion *dvPtr = new DynamicVersion (className, actualName, localName, file, e_TRUE);

	/* load the shared library */

    try {
	loadLibrary (dvPtr, nameMap[handle]);
    } catch (Exception &except) {
	delete (dvPtr);
	throw;
    }

	/* invalidate all non-active versions if we are invalidating; */
	/* otherwise just unmark the active version                   */

    if (invalidate == e_TRUE) {

	invalidateAll (handle, NULL);

    } else if (activeMap[handle] != NULL) {

	deactivateVersion (handle, activeMap[handle], e_FALSE);
    }

	/* mark the new version as the active version */

    versionList.insert (versionList.begin(), dvPtr);
    activeMap[handle] = dvPtr;
    return (handle);
}

template<class T> int dynmedium<T>::activate (const char *actualName, const char *className) THROW5(NoTemporaryLibrary,NoDownload,NoVersionLibrary,InvalidVersion,NoCreateFunction)
{
    return (activateInternal (e_FALSE, actualName, className));
}

/* template <class T> dynmedium<T>::activate_and_invalidate

   Purpose: Register a class version as the current active version *and*
	    invalidate all previous versions

     Input: handle     = the class handle
			 (int)

	    actualName = actual name of the version library
		         (const char *) 

    Output: The procedure marks the given version as the active version,
	    invalidates all previous versions and returns tbe HANDLE of the
            associated class.
*/

template<class T> void dynmedium<T>::invalidateAll (int handle, const char *actualName)
{
	/* assertions on the parameters */

    assert (actualName != NULL);

	/* invalidate all previous versions */

    VersionList::iterator i;
    VersionList &versionList = *versionMap[handle];

    for (i = versionList.begin(); i != versionList.end(); ++i) {
	if ((*i) -> isLibrary (actualName)) {
	    continue;
	} else {
	    deactivateVersion (handle, *i, e_TRUE);
	}
    }
}

template<class T> int dynmedium<T>::activate_and_invalidate (const char *actualName, const char *className) THROW5(NoTemporaryLibrary,NoDownload,NoVersionLibrary,InvalidVersion,NoCreateFunction)
{
    return (activateInternal (e_TRUE, actualName, className));
}


/* ---------------------------- DEACTIVATION ---------------------------- */

template <class T> void dynmedium<T>::deactivateVersion (int handle, DynamicVersion *ptr, Boolean force) 
{
    if (ptr -> refCount == 0) {
	eraseVersion (handle, ptr);
    } else {
        ptr -> active = e_FALSE; 
	ptr -> force = force;
    }
}

/* template <class T> void dynmedium<T>::invalidate

   Purpose: Turn off a version and disallow continued use of that version,
	    i.e., all further calls to operator -> will fail.

     Input: actualName = actual name of the version library
		         (const char *)

	    handle     = class handle
		         (int)

    Output: The procedure turns off the version and disallows continued use.
*/

template <class T> void dynmedium<T>::invalidate (const char *actualName, int handle) THROW3(InvalidHandle,ActiveVersion,NonexistentVersion)
{
    VersionList::iterator i;

	/* assertions */

    assert (actualName != NULL);
	
	/* lock for mutual exclusion -- automatically released */

    DEFINE_GUARD(guard,lock);

	/* check the handle */

    if ((handle < 0) || (handle > numClasses) || (activeMap[handle] == NULL)) {
	throw InvalidHandle();
    }

	/* find the version and turn it off */

    DynamicVersion *ptr = NULL;
    VersionList &versionList = *versionMap[handle];

    for (i = versionList.begin(); i != versionList.end(); ++i) {
	if ((*i) -> isLibrary (actualName)) { 
	    ptr = (*i);
	    break;
	}
    }

    if (ptr != NULL) {

	if (ptr -> active) {
	    throw ActiveVersion();
	}

	deactivateVersion (handle, ptr, e_TRUE);

    } else {

	throw NonexistentVersion (nameMap[handle], actualName);
    }
}

/* ---------------------- Informational routines ------------------------ */

/* template <class T> dynmedium<T>::getVersionCount

   Purpose: Determine the current number of versions for a particular class

     Input: handle = the class handle
		     (int)

    Output: The procedure returns the current number of versions.
*/

template <class T> dynmedium<T>::getVersionCount (int handle)
{
	/* lock for mutual exclusion -- automatically released */

    DEFINE_GUARD(guard,lock);
    return (versionMap[handle] -> size());
}


/* -------------------------- "Smart" pointer --------------------------- */

/* template <class T> dynmedium<T>::operator->

   Purpose: Get the pointer to the actual object

     Input: None

    Output: The procedure returns a pointer to the actual object.
*/

template <class T> T * dynmedium<T>::operator-> (void) THROW1(ZombieVersion)
{
	/* see if object usage is allowed */ 

    if (dvPtr -> force) {
	throw ZombieVersion();
    }

    return (object);
}

#endif /* _DYNAMIC_H */
