/* Copyright (c) 2003 by Hanbit Media Co. All rights reserved*/
/* Lee, Hae-Young(hae_lee@hanmail.net) */
/* Advanced C++ Programming Styles and Idioms, exercise 5-5 */

#include <iostream.h>
#include <stddef.h>
#include <memory.h>

// LAPD class definition 
//
class LAPD {
friend class LAPDMemoryManager;
public:
    LAPD(const char*);
    void *operator new(size_t);
    void operator  delete(void *); 
    virtual int  size()  { return 0; }
    virtual void print() { cout << "class LAPD instance(size = " << minfo.size << ")" << endl; }

protected:
    LAPD() { /* no-op */ }
    union {
        struct {
          unsigned char flag;
          unsigned int  sapi:6;
	        unsigned int  commandResponse:1;
	        unsigned int  zero:1;
	        unsigned int  tei:7;
	        unsigned int  ext:1;
          unsigned char control;
        } header;
        struct {
            LAPD *linkf, *linkb;      
            unsigned short size;
        } minfo;
    };

    unsigned char  tag;   // buddy system allocated tag
    int performCRCCheck() { /* . . . */ return 0; }
};

// LAPDMemoryManager class definition
//
#define round(a,b) (((a+b-1)/b)*b)

class LAPDMemoryManager {
friend class LAPD;
public:
    LAPDMemoryManager() {
        LAPD* buf = (LAPD*)this->buf;
        
        avail = (LAPD*)availBuf;
        avail[Log2MessageBufSize].minfo.linkf = buf;
        avail[Log2MessageBufSize].minfo.linkb = buf;
        for (int k = 0; k < Log2MessageBufSize; k++) 
            avail[k].minfo.linkf = avail[k].minfo.linkb = (LAPD*)&avail[k];

        buf[0].tag = 1;
        buf[0].minfo.linkf = &avail[Log2MessageBufSize];
        buf[0].minfo.linkb = &avail[Log2MessageBufSize];
        buf[0].minfo.size  = Log2MessageBufSize;
    }
  
    LAPD *largestBlock() {
        for (int k = Log2MessageBufSize; k >= 0; --k) {
            if (avail[k].minfo.linkf != &avail[k]) {
                savej = k;
                return avail[k].minfo.linkf;
            }
        }
        return 0;
    }

    void allocateResizeBlock(int sizeNeeded) {
        int k, j = savej;
        
        // roune up to next 2**k
        for (k = 1; k < Log2MessageBufSize; k++)
            if ((1 << k) >= sizeNeeded) break;
        
        // remove the block from list
        LAPD* l = avail[j].minfo.linkf;
        avail[j].minfo.linkf = l->minfo.linkf;
        l->minfo.linkf->minfo.linkb = &avail[j];
        l->tag = 0;

        // if j == k, we found a block with the right size.
        if (j == k) return;

        // split the block
        for (--j; j - k >= 0; --j) {
            LAPD*  p = (LAPD*)(((char *)l) + (1 << j));
            p->tag   = 1;
            p->minfo.size  = j;
            p->minfo.linkf = &avail[j];
            p->minfo.linkb = &avail[j];
            avail[j].minfo.linkf = p;
            avail[j].minfo.linkb = p;

            l->minfo.size = j;
        }
    }

    void deallocateBlock(LAPD *l, int sizeReturned) {
        // roune up to next 2**k
        int k;
        for (k = 1; k < Log2MessageBufSize; k++)
            if ((1 << k) >= sizeReturned) break;

        // combine with buddy if possible
        for ( ; k < Log2MessageBufSize; k++) {
            // is buddy available?
            LAPD* p = buddy(k, l);
            if (p->tag == 0 || (p->tag == 1 && p->minfo.size != k)) break;
            
            // combine with buddy
            p->minfo.linkb->minfo.linkf = p->minfo.linkf;
            p->minfo.linkf->minfo.linkb = p->minfo.linkb;
            
            if (p < l)  l = p;
        }
        l->tag = 1;
        l->minfo.size  = k;
        l->minfo.linkf = avail[k].minfo.linkf;
        avail[k].minfo.linkf->minfo.linkb = l;
        l->minfo.linkb = &avail[k];
        avail[k].minfo.linkf = l;
    }

private:
    enum { MessageBufSize = 4096, Log2MessageBufSize = 12 };
    unsigned char availBuf[(1+Log2MessageBufSize)*round(sizeof(LAPD),sizeof(LAPD*))];
    unsigned char buf[MessageBufSize];
    LAPD *avail;
    int  savej;
    
    LAPD *buddy(int k, LAPD *l) {
        unsigned char *cp = (unsigned char *) l;
        return (LAPD*)(buf + (long(cp-buf) ^ (1<<k)));
    }
};

static LAPDMemoryManager manager;

// Setup class definition
//
class Setup : public LAPD {
friend class LAPD;
public:
    void print() { cout << "class Setup instance" << endl; }

private:
    struct SetupPacketBody {
        unsigned char rep[512];
    } body;
    
    Setup(const char* m) : LAPD() {
        manager.allocateResizeBlock(size());
        ::memcpy(&body, m + sizeof(header), sizeof(body));
    }
    int size() { return sizeof(Setup); }
};

// Setup class definition
//
class Conn : public LAPD {
friend class LAPD;
public:
    void print() { cout << "class Conn instance" << endl; }

private:
    struct ConnPacketBody {
        unsigned char rep[512];
    } body;
    
    Conn(const char* m) : LAPD() {
        manager.allocateResizeBlock(size());
        ::memcpy(&body, m + sizeof(header), sizeof(body));
    }
    int size() { return sizeof(Conn); }
};

// this operator allows one object in the LAPD
// hierarchy to overlay itself with an instance
// of one of its derived classes

inline void *operator new(size_t, LAPD* l) { return l; }

// LAPD class implementation
//
void* LAPD::operator new(size_t) {
    return manager.largestBlock();
}

void LAPD::operator delete(void *l) {
    manager.deallocateBlock((LAPD*)l, ((LAPD*)l)->size());
}

LAPD::LAPD(const char* bits) {
    performCRCCheck();

    // Determine type of message based on address
    switch(bits[0]) {
    case 's':
        (void) ::new(this) Setup(bits); 
        ::memcpy(&header, bits, sizeof(header));
        break;
    case 'c':
        (void) ::new(this) Conn(bits); break;
        ::memcpy(&header, bits, sizeof(header));
        break;
    default:
        break;
    }
}

int main() {
    LAPD* l = new LAPD("setup message");
    l->print();
    delete l;
    
    LAPD* m = new LAPD("conn message");
    m->print();
    delete m;

    return 0;
}
